<!DOCTYPE html>
<html>
   <body>
      <h1>Hello WAMP/Browser - entry page</h1>
      
      Please open one each of

      <ul>
         <li>
            <a href="frontend.html" target="_blank">Frontend</a>
         </li>
         <li>
            <a href="backend.html" target="_blank">Backend</a>
         </li>
      </ul>

      (right click and open in new tab)
   </body>
</html>
